@@ -44,6 +44,7 @@
 #include <gtk/gtknotebook.h>
 #include <gtk/gtkstock.h>
 #include <gtk/gtkbindings.h>
+#include <gtk/gtkgc.h>
 
 enum {
   MOVE_CURSOR,
@@ -181,6 +182,9 @@ static void     editable_real_set_position    (GtkEditable *editable,
 					       gint         position);
 static gint     editable_get_position         (GtkEditable *editable);
 
+static GdkGC *  make_cursor_gc                (GtkWidget   *widget,
+					       const gchar *property_name,
+					       GdkColor    *fallback);
 
 
 
@@ -886,6 +890,30 @@ eel_editable_label_get_line_wrap (EelEditableLabel *label)
   return label->wrap;
 }
 
+PangoFontDescription *
+eel_editable_label_get_font_description (EelEditableLabel *label)
+{
+  if (label->font_desc)
+    return pango_font_description_copy (label->font_desc);
+
+  return NULL;
+}
+
+void
+eel_editable_label_set_font_description (EelEditableLabel *label,
+					 const PangoFontDescription *desc)
+{
+  if (label->font_desc)
+    pango_font_description_free (label->font_desc);
+
+  if (desc)
+    label->font_desc = pango_font_description_copy (desc);
+  else
+    label->font_desc = NULL;
+
+  eel_editable_label_clear_layout (label);
+}
+
 static void
 eel_editable_label_finalize (GObject *object)
 {
@@ -895,6 +923,12 @@ eel_editable_label_finalize (GObject *object)
   
   label = EEL_EDITABLE_LABEL (object);
 
+  if (label->font_desc)
+    {
+      pango_font_description_free (label->font_desc);
+      label->font_desc = NULL;
+    }
+  
   g_object_unref (G_OBJECT (label->im_context));
   label->im_context = NULL;
   
@@ -1014,7 +1048,7 @@ eel_editable_label_ensure_layout (EelEditableLabel *label,
 	  g_string_prepend_len (tmp_string, label->text, label->n_bytes);
 	  g_string_insert (tmp_string, label->selection_anchor, preedit_string);
       
-	  label->layout = gtk_widget_create_pango_layout (widget, label->text);
+	  label->layout = gtk_widget_create_pango_layout (widget, tmp_string->str);
       
 	  pango_attr_list_splice (tmp_attrs, preedit_attrs,
 				  label->selection_anchor, preedit_length);
@@ -1026,6 +1060,9 @@ eel_editable_label_ensure_layout (EelEditableLabel *label,
 	  label->layout = gtk_widget_create_pango_layout (widget, label->text);
 	}
       label->layout_includes_preedit = include_preedit;
+
+      if (label->font_desc != NULL)
+	pango_layout_set_font_description (label->layout, label->font_desc);
       
       pango_layout_set_attributes (label->layout, tmp_attrs);
       
@@ -1059,13 +1096,13 @@ eel_editable_label_ensure_layout (EelEditableLabel *label,
 
       if (label->wrap)
 	{
-	  GtkWidgetAuxInfo *aux_info;
 	  gint longest_paragraph;
 	  gint width, height;
-	  
-	  aux_info = _gtk_widget_get_aux_info (widget, FALSE);
-	  if (aux_info && aux_info->width > 0)
-	    pango_layout_set_width (label->layout, aux_info->width * PANGO_SCALE);
+	  gint set_width;
+
+	  gtk_widget_get_size_request (widget, &set_width, NULL);
+	  if (set_width > 0)
+	    pango_layout_set_width (label->layout, set_width * PANGO_SCALE);
 	  else
 	    {
 	      gint wrap_width;
@@ -1136,7 +1173,7 @@ eel_editable_label_size_request (GtkWidget      *widget,
   EelEditableLabel *label;
   gint width, height;
   PangoRectangle logical_rect;
-  GtkWidgetAuxInfo *aux_info;
+  gint set_width;
 
   g_return_if_fail (EEL_IS_EDITABLE_LABEL (widget));
   g_return_if_fail (requisition != NULL);
@@ -1166,9 +1203,9 @@ eel_editable_label_size_request (GtkWidget      *widget,
 
   pango_layout_get_extents (label->layout, NULL, &logical_rect);
   
-  aux_info = _gtk_widget_get_aux_info (widget, FALSE);
-  if (label->wrap && aux_info && aux_info->width > 0)
-    width += aux_info->width;
+  gtk_widget_get_size_request (widget, &set_width, NULL);
+  if (label->wrap && set_width > 0)
+    width += set_width;
   else 
     width += PANGO_PIXELS (logical_rect.width);
   
@@ -1284,6 +1321,95 @@ eel_editable_label_get_cursor_pos (EelEditableLabel  *label,
   pango_layout_get_cursor_pos (label->layout, index, strong_pos, weak_pos);
 }
 
+/* These functions are copies from gtk+, as they are not exported from gtk+ */
+
+static GdkGC *
+make_cursor_gc (GtkWidget   *widget,
+		const gchar *property_name,
+		GdkColor    *fallback)
+{
+  GdkGCValues gc_values;
+  GdkGCValuesMask gc_values_mask;
+  GdkColor *cursor_color;
+
+  gtk_widget_style_get (widget, property_name, &cursor_color, NULL);
+  
+  gc_values_mask = GDK_GC_FOREGROUND;
+  if (cursor_color)
+    {
+      gc_values.foreground = *cursor_color;
+      gdk_color_free (cursor_color);
+    }
+  else
+    gc_values.foreground = *fallback;
+  
+  gdk_rgb_find_color (widget->style->colormap, &gc_values.foreground);
+  return gtk_gc_get (widget->style->depth, widget->style->colormap, &gc_values, gc_values_mask);
+}
+
+static void
+_eel_draw_insertion_cursor (GtkWidget        *widget,
+			    GdkDrawable      *drawable,
+			    GdkGC            *gc,
+			    GdkRectangle     *location,
+                            GtkTextDirection  direction,
+                            gboolean          draw_arrow)
+{
+  gint stem_width;
+  gint arrow_width;
+  gint x, y;
+  gint i;
+  gfloat cursor_aspect_ratio;
+  gint offset;
+  
+  g_return_if_fail (direction != GTK_TEXT_DIR_NONE);
+  
+  gtk_widget_style_get (widget, "cursor-aspect-ratio", &cursor_aspect_ratio, NULL);
+  
+  stem_width = location->height * cursor_aspect_ratio + 1;
+  arrow_width = stem_width + 1;
+
+  /* put (stem_width % 2) on the proper side of the cursor */
+  if (direction == GTK_TEXT_DIR_LTR)
+    offset = stem_width / 2;
+  else
+    offset = stem_width - stem_width / 2;
+  
+  for (i = 0; i < stem_width; i++)
+    gdk_draw_line (drawable, gc,
+		   location->x + i - offset, location->y,
+		   location->x + i - offset, location->y + location->height - 1);
+
+  if (draw_arrow)
+    {
+      if (direction == GTK_TEXT_DIR_RTL)
+        {
+          x = location->x - offset - 1;
+          y = location->y + location->height - arrow_width * 2 - arrow_width + 1;
+  
+          for (i = 0; i < arrow_width; i++)
+            {
+              gdk_draw_line (drawable, gc,
+                             x, y + i + 1,
+                             x, y + 2 * arrow_width - i - 1);
+              x --;
+            }
+        }
+      else if (direction == GTK_TEXT_DIR_LTR)
+        {
+          x = location->x + stem_width - offset;
+          y = location->y + location->height - arrow_width * 2 - arrow_width + 1;
+  
+          for (i = 0; i < arrow_width; i++) 
+            {
+              gdk_draw_line (drawable, gc,
+                             x, y + i + 1,
+                             x, y + 2 * arrow_width - i - 1);
+              x++;
+            }
+        }
+    }
+}
 
 static void
 eel_editable_label_draw_cursor (EelEditableLabel  *label, gint xoffset, gint yoffset)
@@ -1301,7 +1427,6 @@ eel_editable_label_draw_cursor (EelEditableLabel  *label, gint xoffset, gint yof
       GdkRectangle cursor_location;
       GtkTextDirection dir1 = GTK_TEXT_DIR_NONE;
       GtkTextDirection dir2 = GTK_TEXT_DIR_NONE;
-      GdkGC *gc;
 
       keymap_direction =
 	(gdk_keymap_get_direction (gdk_keymap_get_default ()) == PANGO_DIRECTION_LTR) ?
@@ -1341,11 +1466,10 @@ eel_editable_label_draw_cursor (EelEditableLabel  *label, gint xoffset, gint yof
       cursor_location.width = 0;
       cursor_location.height = PANGO_PIXELS (cursor1->height);
 
-      gc = _gtk_get_insertion_cursor_gc (widget, TRUE);
-      _gtk_draw_insertion_cursor (widget, widget->window, gc,
+      _eel_draw_insertion_cursor (widget, widget->window,
+				  label->primary_cursor_gc,
 				  &cursor_location, dir1,
                                   dir2 != GTK_TEXT_DIR_NONE);
-      g_object_unref (gc);
       
       if (dir2 != GTK_TEXT_DIR_NONE)
 	{
@@ -1354,10 +1478,9 @@ eel_editable_label_draw_cursor (EelEditableLabel  *label, gint xoffset, gint yof
 	  cursor_location.width = 0;
 	  cursor_location.height = PANGO_PIXELS (cursor2->height);
 
-	  gc = _gtk_get_insertion_cursor_gc (widget, FALSE);
-	  _gtk_draw_insertion_cursor (widget, widget->window, gc,
+	  _eel_draw_insertion_cursor (widget, widget->window,
+				      label->secondary_cursor_gc,
 				      &cursor_location, dir2, TRUE);
-	  g_object_unref (gc);
 	}
     }
 }
@@ -1378,7 +1501,7 @@ eel_editable_label_expose (GtkWidget      *widget,
   eel_editable_label_ensure_layout (label, TRUE);
   
   if (GTK_WIDGET_VISIBLE (widget) && GTK_WIDGET_MAPPED (widget) &&
-      label->text && (*label->text != '\0'))
+      label->text)
     {
       get_layout_location (label, &x, &y);
       
@@ -1464,6 +1587,7 @@ eel_editable_label_realize (GtkWidget *widget)
   EelEditableLabel *label;
   GdkWindowAttr attributes;
   gint attributes_mask;
+  static GdkColor gray = { 0, 0x8888, 0x8888, 0x8888 };
 
   GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
   label = EEL_EDITABLE_LABEL (widget);
@@ -1501,6 +1625,14 @@ eel_editable_label_realize (GtkWidget *widget)
   gdk_window_set_background (widget->window, &widget->style->base[GTK_WIDGET_STATE (widget)]);
 
   gtk_im_context_set_client_window (label->im_context, widget->window);
+
+  label->primary_cursor_gc = make_cursor_gc (widget,
+					     "cursor-color",
+					     &widget->style->black);
+      
+  label->secondary_cursor_gc = make_cursor_gc (widget,
+					       "secondary-cursor-color",
+					       &gray);
 }
 
 static void
@@ -1510,6 +1642,12 @@ eel_editable_label_unrealize (GtkWidget *widget)
 
   label = EEL_EDITABLE_LABEL (widget);
 
+  gtk_gc_release (label->primary_cursor_gc);
+  label->primary_cursor_gc = NULL;
+  
+  gtk_gc_release (label->secondary_cursor_gc);
+  label->secondary_cursor_gc = NULL;
+
   /* Strange. Copied from GtkEntry, should be NULL? */
   gtk_im_context_set_client_window (label->im_context, NULL);
   
@@ -2576,6 +2714,9 @@ eel_editable_label_move_cursor (EelEditableLabel    *label,
 	case GTK_MOVEMENT_PARAGRAPHS:
 	case GTK_MOVEMENT_PAGES:
 	  break;
+	default:
+	  g_assert_not_reached ();
+	  break;
 	}
     }
   else
@@ -2613,6 +2754,9 @@ eel_editable_label_move_cursor (EelEditableLabel    *label,
 	case GTK_MOVEMENT_PARAGRAPHS:
 	case GTK_MOVEMENT_PAGES:
 	  break;
+	default:
+	  g_assert_not_reached ();
+	  break;
 	}
     }
 
